@@ -228,7 +228,7 @@ static void set_bredr_commands(struct btdev *btdev)
 	btdev->commands[6]  |= 0x01;	/* Set Event Filter */
 	btdev->commands[6]  |= 0x20;	/* Read Stored Link Key */
 	btdev->commands[6]  |= 0x40;	/* Write Stored Link Key */
-	btdev->commands[6]  |= 0x80;	/* Write Stored Link Key */
+	btdev->commands[6]  |= 0x80;	/* Delete Stored Link Key */
 	btdev->commands[7]  |= 0x01;	/* Write Local Name */
 	btdev->commands[7]  |= 0x02;	/* Read Local Name */
 	btdev->commands[7]  |= 0x04;	/* Read Connection Accept Timeout */
@@ -583,6 +583,7 @@ static void inquiry_complete(struct btdev *btdev, uint8_t status)
 			ir.pscan_rep_mode = 0x00;
 			ir.pscan_period_mode = 0x00;
 			memcpy(ir.dev_class, btdev_list[i]->dev_class, 3);
+			ir.clock_offset = 0x0000;
 			ir.rssi = -60;
 			memcpy(ir.data, btdev_list[i]->ext_inquiry_rsp, 240);
 
@@ -599,6 +600,7 @@ static void inquiry_complete(struct btdev *btdev, uint8_t status)
 			ir.pscan_rep_mode = 0x00;
 			ir.pscan_period_mode = 0x00;
 			memcpy(ir.dev_class, btdev_list[i]->dev_class, 3);
+			ir.clock_offset = 0x0000;
 			ir.rssi = -60;
 
 			send_event(btdev, BT_HCI_EVT_INQUIRY_RESULT_WITH_RSSI,
@@ -612,6 +614,7 @@ static void inquiry_complete(struct btdev *btdev, uint8_t status)
 			ir.pscan_period_mode = 0x00;
 			ir.pscan_mode = 0x00;
 			memcpy(ir.dev_class, btdev_list[i]->dev_class, 3);
+			ir.clock_offset = 0x0000;
 
 			send_event(btdev, BT_HCI_EVT_INQUIRY_RESULT,
 							&ir, sizeof(ir));
@@ -662,7 +665,7 @@ static void conn_request(struct btdev *btdev, const uint8_t *bdaddr)
 	struct btdev *remote = find_btdev_by_bdaddr(bdaddr);
 
 	if (remote) {
-		if (remote->scan_enable & 0x01) {
+		if (remote->scan_enable & 0x02) {
 			struct bt_hci_evt_conn_request cr;
 
 			memcpy(cr.bdaddr, btdev->bdaddr, 6);
@@ -871,6 +874,7 @@ static void default_cmd(struct btdev *btdev, uint16_t opcode,
 	struct bt_hci_rsp_le_read_adv_tx_power lratp;
 	struct bt_hci_rsp_le_read_supported_states lrss;
 	struct bt_hci_rsp_le_read_white_list_size lrwls;
+	struct bt_hci_rsp_remote_name_request_cancel rnrc_rsp;
 	uint8_t status, page;
 
 	switch (opcode) {
@@ -938,8 +942,9 @@ static void default_cmd(struct btdev *btdev, uint16_t opcode,
 		if (btdev->type == BTDEV_TYPE_LE)
 			goto unsupported;
 		rnrc = data;
-		status = BT_HCI_ERR_SUCCESS;
-		cmd_complete(btdev, opcode, &status, sizeof(status));
+		rnrc_rsp.status = BT_HCI_ERR_SUCCESS;
+		memcpy(rnrc_rsp.bdaddr, rnrc->bdaddr, 6);
+		cmd_complete(btdev, opcode, &rnrc_rsp, sizeof(rnrc_rsp));
 		name_request_complete(btdev, rnrc->bdaddr,
 						BT_HCI_ERR_UNKNOWN_CONN_ID);
 		break;
