@@ -29,6 +29,11 @@
 #define BT_H4_SCO_PKT	0x03
 #define BT_H4_EVT_PKT	0x04
 
+struct bt_hci_acl_hdr {
+	uint16_t handle;
+	uint16_t dlen;
+} __attribute__ ((packed));
+
 struct bt_hci_cmd_hdr {
 	uint16_t opcode;
 	uint8_t	 plen;
@@ -162,6 +167,10 @@ struct bt_hci_cmd_remote_name_request {
 struct bt_hci_cmd_remote_name_request_cancel {
 	uint8_t  bdaddr[6];
 } __attribute__ ((packed));
+struct bt_hci_rsp_remote_name_request_cancel {
+	uint8_t  status;
+	uint8_t  bdaddr[6];
+} __attribute__ ((packed));
 
 #define BT_HCI_CMD_READ_REMOTE_FEATURES		0x041b
 struct bt_hci_cmd_read_remote_features {
@@ -690,6 +699,14 @@ struct bt_hci_cmd_write_le_host_supported {
 	uint8_t  simultaneous;
 } __attribute__ ((packed));
 
+#define BT_HCI_CMD_READ_SYNC_TRAIN_PARAMS	0x0c77
+struct bt_hci_rsp_read_sync_train_params {
+	uint8_t  status;
+	uint16_t interval;
+	uint32_t timeout;
+	uint8_t  service_data;
+} __attribute__ ((packed));
+
 #define BT_HCI_CMD_READ_LOCAL_VERSION		0x1001
 struct bt_hci_rsp_read_local_version {
 	uint8_t  status;
@@ -956,6 +973,25 @@ struct bt_hci_cmd_le_start_encrypt {
 	uint8_t  ltk[16];
 } __attribute__ ((packed));
 
+#define BT_HCI_CMD_LE_LTK_REQ_REPLY		0x201a
+struct bt_hci_cmd_le_ltk_req_reply {
+	uint16_t handle;
+	uint8_t  ltk[16];
+} __attribute__ ((packed));
+struct bt_hci_cmd_le_ltk_req_reply_rsp {
+	uint8_t  status;
+	uint16_t handle;
+} __attribute__ ((packed));
+
+#define BT_HCI_CMD_LE_LTK_REQ_NEG_REPLY		0x201b
+struct bt_hci_cmd_le_ltk_req_neg_reply {
+	uint16_t handle;
+} __attribute__ ((packed));
+struct bt_hci_cmd_le_ltk_req_neg_reply_rsp {
+	uint8_t  status;
+	uint16_t handle;
+} __attribute__ ((packed));
+
 #define BT_HCI_CMD_LE_READ_SUPPORTED_STATES	0x201c
 struct bt_hci_rsp_le_read_supported_states {
 	uint8_t  status;
@@ -1483,7 +1519,7 @@ struct bt_l2cap_pdu_config_req {
 
 #define BT_L2CAP_PDU_CONFIG_RSP		0x05
 struct bt_l2cap_pdu_config_rsp {
-	uint16_t dcid;
+	uint16_t scid;
 	uint16_t flags;
 	uint16_t result;
 } __attribute__ ((packed));
