@@ -75,7 +75,7 @@ struct obc_transfer {
 	char *session;		/* Session path */
 	char *owner;		/* Transfer initiator */
 	char *path;		/* Transfer path */
-	gchar *filename;	/* Transfer file location */
+	char *filename;		/* Transfer file location */
 	char *name;		/* Transfer object name */
 	char *type;		/* Transfer object type */
 	int fd;
@@ -144,7 +144,7 @@ static DBusMessage *obc_transfer_cancel(DBusConnection *connection,
 					DBusMessage *message, void *user_data)
 {
 	struct obc_transfer *transfer = user_data;
-	const gchar *sender;
+	const char *sender;
 
 	sender = dbus_message_get_sender(message);
 	if (g_strcmp0(transfer->owner, sender) != 0)
@@ -327,7 +327,8 @@ static void obc_transfer_free(struct obc_transfer *transfer)
 	}
 
 	if (transfer->op == G_OBEX_OP_GET &&
-					transfer->transferred != transfer->size)
+				transfer->status != TRANSFER_STATUS_COMPLETE &&
+				transfer->filename)
 		remove(transfer->filename);
 
 	if (transfer->fd > 0)
@@ -535,7 +536,7 @@ static gboolean get_xfer_progress(const void *buf, gsize len,
 	struct obc_transfer *transfer = user_data;
 
 	if (transfer->fd > 0) {
-		gint w;
+		int w;
 
 		w = write(transfer->fd, buf, len);
 		if (w < 0)
